
# ============================================================================
# AWS Bedrock Error Mapping
# Docs: https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Converse.html#API_runtime_Converse_Errors
# ============================================================================

from llmfy.exception.llmfy_exception import AuthenticationException, InvalidRequestException, ModelErrorException, ModelNotFoundException, PermissionDeniedException, RateLimitException, ServiceUnavailableException, TimeoutException


BEDROCK_ERROR_MAP = {
    # ThrottlingException - HTTP 429
    'ThrottlingException': (RateLimitException, 429),
    
    # ModelTimeoutException - HTTP 408
    'ModelTimeoutException': (TimeoutException, 408),
    
    # ModelNotReadyException - HTTP 429
    'ModelNotReadyException': (ServiceUnavailableException, 429),
    
    # ValidationException - HTTP 400
    'ValidationException': (InvalidRequestException, 400),
    
    # AccessDeniedException - HTTP 403
    'AccessDeniedException': (AuthenticationException, 403),
    
    # ResourceNotFoundException - HTTP 404
    'ResourceNotFoundException': (ModelNotFoundException, 404),
    
    # ServiceUnavailableException - HTTP 503
    'ServiceUnavailableException': (ServiceUnavailableException, 503),
    
    # InternalServerException - HTTP 500
    'InternalServerException': (ServiceUnavailableException, 500),
    
    # ModelErrorException - HTTP 424
    'ModelErrorException': (ModelErrorException, 424),
}


# ============================================================================
# OpenAI Error Mapping
# Docs: https://github.com/openai/openai-python#handling-errors
# ============================================================================

OPENAI_ERROR_MAP = {
    'RateLimitError': (RateLimitException, 429),
    'APITimeoutError': (TimeoutException, 408),
    'APIConnectionError': (ServiceUnavailableException, None),
    'AuthenticationError': (AuthenticationException, 401),
    'PermissionDeniedError': (PermissionDeniedException, 403),
    'BadRequestError': (InvalidRequestException, 400),
    'NotFoundError': (ModelNotFoundException, 404),
    'UnprocessableEntityError': (InvalidRequestException, 422),
    'InternalServerError': (ServiceUnavailableException, 500),
}


# ============================================================================
# Google Gen AI Error Mapping
# Docs: https://github.com/googleapis/python-genai#error-handling
# ============================================================================

GOOGLE_ERROR_MAP = {
    429: RateLimitException,
    408: TimeoutException,
    400: InvalidRequestException,
    401: AuthenticationException,
    403: PermissionDeniedException,
    404: ModelNotFoundException,
    500: ServiceUnavailableException,
    503: ServiceUnavailableException,
}





from .embeddings.base_embedding_model import BaseEmbeddingModel
from .embeddings.bedrock.bedrock_embedding import BedrockEmbedding
from .embeddings.openai.openai_embedding import OpenAIEmbedding
from .llmfy import LLMfy
from .messages import Content, ContentType, Message, MessageTemp, Role, ToolCall
from .models import (
    BEDROCK_PRICING,
    OPENAI_PRICING,
    BaseAIModel,
    BedrockConfig,
    BedrockFormatter,
    BedrockModel,
    ModelPricing,
    OpenAIConfig,
    OpenAIModel,
)
from .responses import AIResponse, GenerationResponse
from .tools import Tool, ToolRegistry
from .usage import LLMfyUsage, llmfy_usage_tracker

__all__ = [
    "LLMfy",
    "MessageTemp",
    "Message",
    "Role",
    "ToolCall",
    "ToolRegistry",
    "Tool",
    "AIResponse",
    "GenerationResponse",
    "BaseAIModel",
    "ModelPricing",
    "OpenAIConfig",
    "OpenAIModel",
    "OPENAI_PRICING",
    "BedrockConfig",
    "BedrockFormatter",
    "BedrockModel",
    "BEDROCK_PRICING",
    "Content",
    "ContentType",
    "llmfy_usage_tracker",
    "LLMfyUsage",
    "BaseEmbeddingModel",
    "BedrockEmbedding",
    "OpenAIEmbedding",
]



import asyncio
from typing import List, Optional

from sqlalchemy import TypeDecorator
from sqlalchemy.dialects import mysql, postgresql

from llmfy.exception.llmfy_exception import LLMfyException
from llmfy.flow_engine.checkpointer.base_checkpointer import (
    BaseCheckpointer,
    Checkpoint,
    CheckpointMetadata,
)

try:
    from sqlalchemy import (
        Column,
        DateTime,
        Index,
        Integer,
        String,
        Text,
        create_engine,
        delete,
        select,
    )
    from sqlalchemy.ext.asyncio import (
        AsyncSession,
        async_sessionmaker,
        create_async_engine,
    )
    from sqlalchemy.orm import declarative_base, sessionmaker

    SQLALCHEMY_AVAILABLE = True
except ImportError:
    SQLALCHEMY_AVAILABLE = False


class TimestampMilliseconds(TypeDecorator):
    impl = DateTime
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "mysql":
            return dialect.type_descriptor(mysql.DATETIME(fsp=3))
        elif dialect.name == "postgresql":
            return dialect.type_descriptor(postgresql.TIMESTAMP(precision=3))
        else:
            return dialect.type_descriptor(DateTime())


class LongText(TypeDecorator):
    impl = Text
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(postgresql.TEXT())
        elif dialect.name == "mysql":
            return dialect.type_descriptor(mysql.LONGTEXT())
        else:
            return dialect.type_descriptor(Text())


if SQLALCHEMY_AVAILABLE:
    Base = declarative_base()

    class CheckpointModel(Base):
        """SQLAlchemy model for checkpoint storage."""

        __tablename__ = "llmfy_checkpoint"

        checkpoint_id = Column(String(255), primary_key=True)
        session_id = Column(String(255), nullable=False, index=True)
        timestamp = Column(TimestampMilliseconds, nullable=False)
        node_name = Column(String(255), nullable=False)
        step = Column(Integer, nullable=False)
        state = Column(LongText, nullable=False)

        # Composite index for efficient queries
        __table_args__ = (Index("idx_thread_timestamp", "session_id", "timestamp"),)

else:
    raise LLMfyException(
        "SQLAlchemy package is not installed. SQLAlchemy package is required for SQLCheckpointer. "
        'Install it using `pip install "llmfy[SQLAlchemy]"`'
    )


class SQLCheckpointer(BaseCheckpointer):
    """
    SQL database checkpoint storage backend using SQLAlchemy.

    Supports both sync and async drivers for multiple databases:
    - PostgreSQL (async: asyncpg, sync: psycopg2)
    - MySQL (async: aiomysql, sync: pymysql)
    - SQLite (async: aiosqlite, sync: built-in)
    """

    def __init__(self, connection_string: str, echo: bool = False):
        """
        Initialize the SQL database checkpointer.

        Args:
            connection_string: SQLAlchemy connection string (sync or async)
            echo: Whether to echo SQL statements (for debugging)

        Example connection strings:

            ASYNC (Recommended):
            - PostgreSQL: "postgresql+asyncpg://user:password@localhost:5432/dbname"
            - MySQL:      "mysql+aiomysql://user:password@localhost:3306/dbname"
            - SQLite:     "sqlite+aiosqlite:///./database.db"

            SYNC (For compatibility with pymysql, psycopg2, etc):
            - PostgreSQL: "postgresql+psycopg2://user:password@localhost:5432/dbname"
            - MySQL:      "mysql+pymysql://user:password@localhost:3306/dbname"
            - SQLite:     "sqlite:///./database.db"

        Installation:
            Async drivers (recommended):
            - pip install sqlalchemy asyncpg --break-system-packages      # PostgreSQL
            - pip install sqlalchemy aiomysql --break-system-packages     # MySQL
            - pip install sqlalchemy aiosqlite --break-system-packages    # SQLite

            Sync drivers (for compatibility):
            - pip install sqlalchemy psycopg2-binary --break-system-packages  # PostgreSQL
            - pip install sqlalchemy pymysql --break-system-packages          # MySQL
            - pip install sqlalchemy --break-system-packages                  # SQLite (built-in)
        """
        if not SQLALCHEMY_AVAILABLE:
            raise ImportError(
                "SQLAlchemy is required for SQLCheckpointer.\n"
                "Install with: pip install sqlalchemy --break-system-packages\n\n"
                "Then install your database driver:\n"
                "  Async (recommended):\n"
                "    pip install asyncpg --break-system-packages      # PostgreSQL\n"
                "    pip install aiomysql --break-system-packages     # MySQL\n"
                "    pip install aiosqlite --break-system-packages    # SQLite\n\n"
                "  Sync (for compatibility):\n"
                "    pip install psycopg2-binary --break-system-packages  # PostgreSQL\n"
                "    pip install pymysql --break-system-packages          # MySQL\n"
                "    (SQLite is built-in, no extra package needed)"
            )

        self.connection_string = connection_string

        # Detect if using async or sync driver
        self.is_async = any(
            driver in connection_string.lower()
            for driver in ["asyncpg", "aiomysql", "aiosqlite", "+async"]
        )

        if self.is_async:
            # Async mode
            self.engine = create_async_engine(connection_string, echo=echo)
            self.session_maker = async_sessionmaker(
                self.engine, class_=AsyncSession, expire_on_commit=False
            )
        else:
            # Sync mode
            self.engine = create_engine(connection_string, echo=echo)
            self.session_maker = sessionmaker(bind=self.engine)

        self._initialized = False

    async def _ensure_initialized(self):
        """Ensure database tables are created."""
        if not self._initialized:
            if self.is_async:
                async with self.engine.begin() as conn:  # type: ignore
                    await conn.run_sync(Base.metadata.create_all)
            else:
                # Sync initialization - run in executor to keep it async
                loop = asyncio.get_event_loop()
                await loop.run_in_executor(None, Base.metadata.create_all, self.engine)

            self._initialized = True

    async def save(self, checkpoint: Checkpoint) -> None:
        """Save a checkpoint to SQL database."""
        await self._ensure_initialized()

        model = CheckpointModel(
            checkpoint_id=checkpoint.metadata.checkpoint_id,
            session_id=checkpoint.metadata.session_id,
            timestamp=checkpoint.metadata.timestamp,
            node_name=checkpoint.metadata.node_name,
            step=checkpoint.metadata.step,
            state=Checkpoint._serialize_state(checkpoint.state),
        )

        if self.is_async:
            async with self.session_maker() as session:  # type: ignore
                session.add(model)
                await session.commit()
        else:
            # Sync operation - run in executor
            loop = asyncio.get_event_loop()
            await loop.run_in_executor(None, self._save_sync, model)

    def _save_sync(self, model: CheckpointModel):
        """Helper for sync save."""
        with self.session_maker() as session:  # type: ignore
            session.add(model)
            session.commit()

    async def load(
        self,
        session_id: str,
        checkpoint_id: Optional[str] = None,
    ) -> Optional[Checkpoint]:
        """Load a checkpoint from SQL database."""
        await self._ensure_initialized()

        if self.is_async:
            return await self._load_async(session_id, checkpoint_id)
        else:
            # Sync operation - run in executor
            loop = asyncio.get_event_loop()
            return await loop.run_in_executor(
                None, self._load_sync, session_id, checkpoint_id
            )

    async def _load_async(
        self, session_id: str, checkpoint_id: Optional[str]
    ) -> Optional[Checkpoint]:
        """Helper for async load."""
        async with self.session_maker() as session:  # type: ignore
            if checkpoint_id:
                stmt = select(CheckpointModel).where(
                    CheckpointModel.checkpoint_id == checkpoint_id,
                    CheckpointModel.session_id == session_id,
                )
            else:
                stmt = (
                    select(CheckpointModel)
                    .where(CheckpointModel.session_id == session_id)
                    .order_by(CheckpointModel.timestamp.desc())
                    .limit(1)
                )

            result = await session.execute(stmt)
            model = result.scalar_one_or_none()

            return self._model_to_checkpoint(model) if model else None

    def _load_sync(
        self,
        session_id: str,
        checkpoint_id: Optional[str],
    ) -> Optional[Checkpoint]:
        """Helper for sync load."""
        with self.session_maker() as session:  # type: ignore
            if checkpoint_id:
                stmt = select(CheckpointModel).where(
                    CheckpointModel.checkpoint_id == checkpoint_id,
                    CheckpointModel.session_id == session_id,
                )
            else:
                stmt = (
                    select(CheckpointModel)
                    .where(CheckpointModel.session_id == session_id)
                    .order_by(CheckpointModel.timestamp.desc())
                    .limit(1)
                )

            result = session.execute(stmt)
            model = result.scalar_one_or_none()

            return self._model_to_checkpoint(model) if model else None

    async def list(self, session_id: str, limit: int = 10) -> List[Checkpoint]:
        """List checkpoints for a session."""
        await self._ensure_initialized()

        if self.is_async:
            return await self._list_async(session_id, limit)
        else:
            loop = asyncio.get_event_loop()
            return await loop.run_in_executor(None, self._list_sync, session_id, limit)

    async def _list_async(self, session_id: str, limit: int) -> List[Checkpoint]:
        """Helper for async list."""
        async with self.session_maker() as session:  # type: ignore
            stmt = (
                select(CheckpointModel)
                .where(CheckpointModel.session_id == session_id)
                .order_by(CheckpointModel.timestamp.desc())
                .limit(limit)
            )
            result = await session.execute(stmt)
            models = result.scalars().all()

            return [self._model_to_checkpoint(model) for model in models]

    def _list_sync(self, session_id: str, limit: int) -> List[Checkpoint]:
        """Helper for sync list."""
        with self.session_maker() as session:  # type: ignore
            stmt = (
                select(CheckpointModel)
                .where(CheckpointModel.session_id == session_id)
                .order_by(CheckpointModel.timestamp.desc())
                .limit(limit)
            )
            result = session.execute(stmt)
            models = result.scalars().all()

            return [self._model_to_checkpoint(model) for model in models]

    async def delete(self, session_id: str, checkpoint_id: Optional[str] = None) -> None:
        """Delete checkpoint(s) from SQL database."""
        await self._ensure_initialized()

        if self.is_async:
            await self._delete_async(session_id, checkpoint_id)
        else:
            loop = asyncio.get_event_loop()
            await loop.run_in_executor(
                None, self._delete_sync, session_id, checkpoint_id
            )

    async def _delete_async(self, session_id: str, checkpoint_id: Optional[str]):
        """Helper for async delete."""
        async with self.session_maker() as session:  # type: ignore
            if checkpoint_id:
                stmt = delete(CheckpointModel).where(
                    CheckpointModel.checkpoint_id == checkpoint_id,
                    CheckpointModel.session_id == session_id,
                )
            else:
                stmt = delete(CheckpointModel).where(
                    CheckpointModel.session_id == session_id
                )

            await session.execute(stmt)
            await session.commit()

    def _delete_sync(self, session_id: str, checkpoint_id: Optional[str]):
        """Helper for sync delete."""
        with self.session_maker() as session:  # type: ignore
            if checkpoint_id:
                stmt = delete(CheckpointModel).where(
                    CheckpointModel.checkpoint_id == checkpoint_id,
                    CheckpointModel.session_id == session_id,
                )
            else:
                stmt = delete(CheckpointModel).where(
                    CheckpointModel.session_id == session_id
                )

            session.execute(stmt)
            session.commit()

    async def clear_all(self) -> None:
        """Clear all checkpoints from SQL database."""
        await self._ensure_initialized()

        if self.is_async:
            await self._clear_all_async()
        else:
            loop = asyncio.get_event_loop()
            await loop.run_in_executor(None, self._clear_all_sync)

    async def _clear_all_async(self):
        """Helper for async clear_all."""
        async with self.session_maker() as session:  # type: ignore
            stmt = delete(CheckpointModel)
            await session.execute(stmt)
            await session.commit()

    def _clear_all_sync(self):
        """Helper for sync clear_all."""
        with self.session_maker() as session:  # type: ignore
            stmt = delete(CheckpointModel)
            session.execute(stmt)
            session.commit()

    @staticmethod
    def _model_to_checkpoint(model: CheckpointModel) -> Checkpoint:
        """Convert SQLAlchemy model to Checkpoint object."""
        metadata = CheckpointMetadata(
            checkpoint_id=model.checkpoint_id,  # type: ignore
            session_id=model.session_id,  # type: ignore
            timestamp=model.timestamp,  # type: ignore
            node_name=model.node_name,  # type: ignore
            step=model.step,  # type: ignore
        )
        state = Checkpoint._deserialize_state(model.state)  # type: ignore
        return Checkpoint(metadata=metadata, state=state)

    async def close(self) -> None:
        """Close the database connection."""
        if self.is_async:
            await self.engine.dispose()  # type: ignore
        else:
            loop = asyncio.get_event_loop()
            await loop.run_in_executor(None, self.engine.dispose)

    async def __aenter__(self):
        """Async context manager entry."""
        await self._ensure_initialized()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        await self.close()



from .chunk import chunk_markdown_by_header, chunk_text
from .text_preprocessing import (
    clean_text_for_embedding,
)

__all__ = [
    "clean_text_for_embedding",
    "chunk_text",
    "chunk_markdown_by_header",
]



from llmfy.exception.exception_mapper import (
    BEDROCK_ERROR_MAP,
    # GOOGLE_ERROR_MAP,
    OPENAI_ERROR_MAP,
)
from llmfy.exception.llmfy_exception import LLMfyException
from llmfy.llmfy_core.service_provider import ServiceProvider


def handle_bedrock_error(e) -> LLMfyException:
    """
    Handle AWS Bedrock ClientError exceptions.

    Docs: https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Converse.html
    """
    from botocore.exceptions import ClientError

    if not isinstance(e, ClientError):
        return LLMfyException(
            message=str(e),
            raw_error=e,
            provider="bedrock",
        )

    error_code = e.response["Error"]["Code"]
    message = e.response["Error"].get("Message", str(e))
    http_status = e.response.get("ResponseMetadata", {}).get("HTTPStatusCode")

    if error_code in BEDROCK_ERROR_MAP:
        exception_class, default_status = BEDROCK_ERROR_MAP[error_code]
        status_code = http_status or default_status
    else:
        exception_class = LLMfyException
        status_code = http_status

    return exception_class(
        message=message,
        status_code=status_code,
        raw_error=e.response,
        provider=ServiceProvider.BEDROCK,
    )


def handle_openai_error(e) -> LLMfyException:
    """
    Handle OpenAI API exceptions.

    Docs: https://github.com/openai/openai-python#handling-errors
    """
    # import openai

    error_type = type(e).__name__

    # Extract status code from APIStatusError subclasses
    status_code = getattr(e, "status_code", None)

    # Build raw error dict
    raw_error = {
        "type": error_type,
        "message": str(e),
    }

    # Add response if available
    if hasattr(e, "response"):
        raw_error["response"] = e.response

    # Add request_id if available
    if hasattr(e, "request_id"):
        raw_error["request_id"] = e.request_id

    # Add body if available
    if hasattr(e, "body"):
        raw_error["body"] = e.body

    if error_type in OPENAI_ERROR_MAP:
        exception_class, default_status = OPENAI_ERROR_MAP[error_type]
        status_code = status_code or default_status
    else:
        exception_class = LLMfyException

    return exception_class(
        message=str(e),
        status_code=status_code,
        raw_error=raw_error,
        provider=ServiceProvider.OPENAI,
    )


# def handle_google_error(e) -> LLMfyException:
#     """
#     Handle Google Gen AI API exceptions.

#     Docs: https://github.com/googleapis/python-genai#error-handling
#     """
#     # from google.genai import errors

#     if isinstance(e, errors.APIError):
#         status_code = e.code
#         message = e.message

#         raw_error = {
#             'code': e.code,
#             'message': e.message,
#         }

#         # Add any additional attributes
#         if hasattr(e, 'details'):
#             raw_error['details'] = e.details

#         exception_class = GOOGLE_ERROR_MAP.get(status_code, LLMfyException)

#         return exception_class(
#             message=message,
#             status_code=status_code,
#             raw_error=raw_error,
#             provider='google',
#         )

#     return LLMfyException(
#         message=str(e),
#         raw_error={'error': str(e)},
#         provider='google',
#     )



from typing import List
from pydantic import BaseModel, ConfigDict, Field

from llmfy.llmfy_core.messages.message import Message
from llmfy.llmfy_core.responses.ai_response import AIResponse


class GenerationResponse(BaseModel):
    """GenerationResponse Class"""

    model_config = ConfigDict(extra="forbid")
    result: AIResponse
    messages: List[Message] = Field(default_factory=list)



from .bedrock_config import BedrockConfig
from .bedrock_formatter import BedrockFormatter
from .bedrock_model import BedrockModel
from .bedrock_pricing_list import BEDROCK_PRICING

__all__ = [
    "BedrockConfig",
    "BedrockFormatter",
    "BedrockModel",
    "BEDROCK_PRICING",
]



from .checkpointer import (
    BaseCheckpointer,
    InMemoryCheckpointer,
    RedisCheckpointer,
    SQLCheckpointer,
)
from .edge import Edge
from .flow_engine import FlowEngine
from .helper import (
    count_tokens_approximately,
    safe_trim_messages,
    tools_node,
    tools_stream_node,
    trim_messages,
)
from .node import END, START, Node, NodeType
from .state import MemoryManager, WorkflowState
from .stream import (
    FlowEngineStreamResponse,
    FlowEngineStreamType,
    NodeStreamResponse,
    NodeStreamType,
    ToolNodeStreamResponse,
    ToolNodeStreamType,
)
from .visualizer import WorkflowVisualizer

__all__ = [
    "FlowEngine",
    "Edge",
    "Node",
    "NodeType",
    "START",
    "END",
    "WorkflowState",
    "MemoryManager",
    "WorkflowVisualizer",
    "BaseCheckpointer",
    "InMemoryCheckpointer",
    "RedisCheckpointer",
    "SQLCheckpointer",
    "tools_node",
    "tools_stream_node",
    "trim_messages",
    "safe_trim_messages",
    "count_tokens_approximately",
    "FlowEngineStreamResponse",
    "FlowEngineStreamType",
    "NodeStreamResponse",
    "NodeStreamType",
    "ToolNodeStreamResponse",
    "ToolNodeStreamType",
]



import functools
import warnings
from typing import Callable, Optional, Union


def deprecated(
    reason: Optional[str] = None,
    version: Optional[str] = None,
    alternative: Optional[str] = None,
    category: type = DeprecationWarning,
) -> Callable:
    """
    Decorator to mark functions, classes, or context managers as deprecated.

    Args:
        reason: Optional reason for deprecation
        version: Version when the item was deprecated
        alternative: Suggested alternative to use instead
        category: Warning category (default: DeprecationWarning)

    Returns:
        Decorated function, class, or context manager
    """

    def decorator(obj: Union[Callable, type]) -> Union[Callable, type]:
        # Build warning message
        obj_type = "class" if isinstance(obj, type) else "function"
        obj_name = getattr(obj, "__name__", str(obj))

        msg_parts = [f"{obj_type.capitalize()} '{obj_name}' is deprecated"]

        if version:
            msg_parts.append(f"since version {version}")
        if reason:
            msg_parts.append(f"({reason})")
        if alternative:
            msg_parts.append(f"Use '{alternative}' instead")

        warning_msg = ". ".join(msg_parts) + "."

        if isinstance(obj, type):
            # Handle classes
            original_init = obj.__init__

            @functools.wraps(original_init)
            def new_init(self, *args, **kwargs):
                warnings.warn(warning_msg, category=category, stacklevel=2)
                return original_init(self, *args, **kwargs)

            obj.__init__ = new_init
            return obj

        else:
            # Handle functions and context managers
            @functools.wraps(obj)
            def wrapper(*args, **kwargs):
                warnings.warn(warning_msg, category=category, stacklevel=2)
                return obj(*args, **kwargs)

            return wrapper

    return decorator



from abc import ABC, abstractmethod
from typing import Any, List

from llmfy.llmfy_core.service_provider import ServiceProvider


class BaseEmbeddingModel(ABC):
    """BaseEmbeddingModel Abstract"""

    def __init__(self):
        """Model provider."""
        self.provider: ServiceProvider
        self.model: str

    @abstractmethod
    def encode(
        self,
        text: str,
    ) -> List[float]:
        pass

    @abstractmethod
    def encode_batch(
        self,
        texts: List[str] | str,
        batch_size: int = 10,
        max_retries: int = 3,
        retry_delay: float = 1.0,
        show_progress_bar: bool = False,
    ) -> Any:
        pass



import inspect
from typing import Any, Dict, List, Union

from llmfy.exception.llmfy_exception import LLMfyException
from llmfy.llmfy_core.messages.content_type import ContentType
from llmfy.llmfy_core.messages.message import Message
from llmfy.llmfy_core.messages.role import Role
from llmfy.llmfy_core.models.model_formatter import ModelFormatter


class BedrockFormatter(ModelFormatter):
    """BedrockFormatter.

    Args:
        id (str): _description_
        role (Role): _description_
        content (Optional[str], optional): _description_. Defaults to None.
        tool_calls (Optional[List[ToolCall]], optional): _description_. Defaults to None.
        tool_call_id (Optional[str], optional): _description_. Defaults to None.
        name (Optional[str], optional): _description_. Defaults to None.

    Returns:
        dict: _description_
    """

    def format_message(self, message: Message) -> dict:
        """Formats message into Bedrock's message format.

        TextRequest:
        ```
        {
            "role": "user | assistant",
            "content": [
                {
                    "text": "string"
                }
            ]
        }
        ```

        ImageRequest:
        ```
        {
            "role": "user",
            "content": [
                {
                    "image": {
                        "format": "png",
                        "source": {
                            "bytes": "image in bytes"
                        }
                    }
                }
            ]
        }
        # or
        {
            "role": "user",
            "content": [
                {
                    "image": {
                        "format": "png",
                        "source": {
                            "s3Location": {
                                "uri": "s3://amzn-s3-demo-bucket/myImage",
                                "bucketOwner": "111122223333"
                            }
                        }
                    }
                }
            ]
        }
        ```

        ToolRequest:
        ```
        {
            "role": "assistant",
            "content": [
                {
                    "toolUse": {
                        "toolUseId": "tooluse_kZJMlvQmRJ6eAyJE5GIl7Q",
                        "name": "top_song",
                        "input": {
                            "sign": "WZPZ"
                        }
                    }
                }
            ]
        }
        ```

        ToolResultRequest:
        ```
        {
            "role": "user",
            "content": [
                {
                    "toolResult": {
                        "toolUseId": "tooluse_kZJMlvQmRJ6eAyJE5GIl7Q",
                        "content": [
                            {
                                "json": {
                                    "song": "Elemental Hotel",
                                    "artist": "8 Storey Hike"
                                }
                            }
                        ]
                    }
                }
            ]
        }
        ```
        """
        # for tool result (tool message) in bedrock role become `user`
        role = message.role.value if message.role.value != "tool" else "user"
        message_dict: Dict[str, Any] = {
            "role": role,
        }

        if message.content and not message.tool_results and not message.tool_calls:
            if isinstance(message.content, str):
                # content is absolute text
                message_dict["content"] = [{"text": message.content}]
            if isinstance(message.content, List):
                # content can be text, image, document or video
                message_dict["content"] = []
                for c in message.content:
                    if c.type == ContentType.TEXT:
                        #  Content.value value is str.
                        message_dict["content"].append({"text": c.value})

                    if c.type == ContentType.IMAGE:
                        supported_formats = ["gif", "jpeg", "png", "webp"]
                        # check format
                        if not c.format:
                            raise LLMfyException("`format` is required for bedrock.")
                        if c.format not in supported_formats:
                            raise LLMfyException(
                                f"`format` must in {supported_formats}."
                            )

                        # check is use s3
                        if c.use_s3:
                            # Use s3
                            # check bwner if use s3
                            if not c.bucket_owner:
                                raise LLMfyException(
                                    "`bucket_owner` is required if use s3."
                                )

                            # use s3
                            # Content.value value is url to s3 image.
                            message_dict["content"].append(
                                {
                                    "image": {
                                        "format": c.format,
                                        "source": {
                                            "s3Location": {
                                                "uri": c.value,
                                                "bucketOwner": c.bucket_owner,
                                            }
                                        },
                                    }
                                }
                            )
                        else:
                            # Use bytes
                            #  Content.value value is str image bytes.
                            message_dict["content"].append(
                                {
                                    "image": {
                                        "format": c.format,
                                        "source": {"bytes": c.value},
                                    },
                                }
                            )

                    if c.type == ContentType.DOCUMENT:
                        # check filename
                        if not c.filename:
                            raise LLMfyException(
                                "`filename` is required for content type DOCUMENT"
                            )

                        # check is use s3
                        if c.use_s3:
                            # Use s3
                            # check bwner if use s3
                            if not c.bucket_owner:
                                raise LLMfyException(
                                    "`bucket_owner` is required if use s3."
                                )

                            # use s3
                            # Content.value value is url to s3 image.
                            message_dict["content"].append(
                                {
                                    "document": {
                                        "format": "pdf",
                                        "name": c.filename,
                                        "source": {
                                            "s3Location": {
                                                "uri": c.value,
                                                "bucketOwner": c.bucket_owner,
                                            }
                                        },
                                    }
                                }
                            )
                        else:
                            # Use bytes
                            # Content.value value is str pdf bytes.
                            message_dict["content"].append(
                                {
                                    "document": {
                                        "format": "pdf",
                                        "name": c.filename,
                                        "source": {"bytes": c.value},
                                    },
                                }
                            )

                    if c.type == ContentType.VIDEO:
                        supported_formats = [
                            "wmv",
                            "mpg",
                            "mpeg",
                            "three_gp",
                            "flv",
                            "mp4",
                            "mov",
                            "mkv",
                            "webm",
                        ]
                        # check format
                        if not c.format:
                            raise LLMfyException("`format` is required for bedrock.")
                        if c.format not in supported_formats:
                            raise LLMfyException(
                                f"`format` must in {supported_formats}."
                            )

                        # check is use s3
                        if c.use_s3:
                            # Use s3
                            # check bwner if use s3
                            if not c.bucket_owner:
                                raise LLMfyException(
                                    "`bucket_owner` is required if use s3."
                                )

                            # use s3
                            # Content.value value is url to s3 video.
                            message_dict["content"].append(
                                {
                                    "video": {
                                        "format": c.format,
                                        "source": {
                                            "s3Location": {
                                                "uri": c.value,
                                                "bucketOwner": c.bucket_owner,
                                            }
                                        },
                                    }
                                }
                            )
                        else:
                            # Use bytes
                            #  Content.value value is str video bytes.
                            message_dict["content"].append(
                                {
                                    "video": {
                                        "format": c.format,
                                        "source": {"bytes": c.value},
                                    },
                                }
                            )

        if message.tool_results:
            message_dict["content"] = message.tool_results

        if message.tool_calls:
            message_dict["content"] = [
                {
                    "toolUse": {
                        "toolUseId": tool_call.tool_call_id,
                        "name": tool_call.name,
                        "input": tool_call.arguments,
                    }
                }
                for tool_call in message.tool_calls
            ]

        if message.name and not message.role.value == "tool":
            message_dict["name"] = message.name

        return message_dict

    def format_tool_function(
        self, func_metadata: Dict, type_mapping: dict[Any, str]
    ) -> Dict:
        """Formats a function into Bedrock's tool format.

        ```
        {
            "tools": [
                {
                    "toolSpec": {
                        "name": "get_current_weather",
                        "description": "",
                        "inputSchema": {
                            "json": {
                                "type": "object",
                                "properties": {
                                    "location": {
                    ucket_o                    "type": "string",
                                        "description": ""
                                    },
                                    "unit": {
                                        "type": "string",
                                        "description": " (default: celsius)"
                                    }
                                },
                                "required": [
                                    "location",
                                    "unit"
                                ]
                            }
                        }
                    }
                }
            ]
        }
        ```
        """
        metadata = func_metadata

        tool_def = {
            "name": metadata["name"],
            "description": metadata["description"] or metadata["name"],
            "inputSchema": {
                "json": {
                    "type": "object",
                    "properties": {},
                    "required": [],
                }
            },
        }

        for param_name, param in metadata["parameters"].items():
            if param_name == "self":  # Skip 'self' for methods
                continue

            python_type = metadata["type_hints"].get(param_name, param.annotation)
            if hasattr(python_type, "__origin__") and python_type.__origin__ is Union:
                # Extract non-None types from Union
                types = [t for t in python_type.__args__ if t is not type(None)]
                python_type = types[0] if len(types) == 1 else str

            param_type = type_mapping.get(python_type, "string")

            # Extract parameter description
            from llmfy.llmfy_core.tools.function_param_desc_extractor import (
                extract_param_desc,
            )

            docstring = metadata["docstring"]
            param_description = extract_param_desc(param_name, docstring)

            # Extract default value
            param_default = (
                f"(default: {param.default})"
                if param.default != inspect.Parameter.empty
                else ""
            )

            # Add parameter details
            tool_def["inputSchema"]["json"]["properties"][param_name] = {
                "type": param_type,
                "description": param_description
                + (" " if param_default else "")
                + param_default,
            }

            # Add required params
            tool_def["inputSchema"]["json"]["required"].append(param_name)

        return tool_def

    def format_tool_message(
        self,
        messages: List[Message],
        id: str,
        tool_call_id: str,
        name: str,
        result: str,
        request_call_id: str | None = None,
    ) -> List[Message]:
        """
        There are 2 kind tool message provided in bedrock,
        - if all tool request in one item list, the tool message also must in one item list, see sample : `app/llmfy/messages/sample_v1_bedrock_messages.json`
        - if tool request is separated one by one, the tool message also must provided one by one, see sample : `app/llmfy/messages/sample_v2_bedrock_messages.json`
        """

        tool_result = {
            "toolResult": {
                "toolUseId": tool_call_id,
                "content": [{"json": {"result": result}}],
            }
        }

        # Find existing Bedrock tool response message
        bedrock_message = next(
            (
                msg
                for msg in messages
                if msg.role == Role.TOOL
                and msg.tool_results
                and msg.request_call_id == request_call_id
            ),
            None,
        )

        if bedrock_message:
            # V1
            # update
            if bedrock_message.tool_results:
                bedrock_message.tool_results.append(tool_result)
        else:
            # V2
            # add new
            messages.append(
                Message(
                    id=id,
                    role=Role.TOOL,
                    tool_results=[tool_result],
                    request_call_id=request_call_id,
                )
            )
        return messages



from .document import Document
from .faiss_index.faiss_index import FAISSIndex
from .faiss_index.faiss_vector_store import FAISSVectorStore

__all__ = [
    "Document",
    "FAISSIndex",
    "FAISSVectorStore",
]



from .messages_trimmer import (
    count_tokens_approximately,
    safe_trim_messages,
    trim_messages,
)
from .tools_node import tools_node, tools_stream_node

__all__ = [
    "tools_node",
    "tools_stream_node",
    "trim_messages",
    "safe_trim_messages",
    "count_tokens_approximately",
]



import base64
import inspect
from typing import Any, AsyncGenerator, Callable, Dict, List, Optional, Union

from llmfy.exception.llmfy_exception import LLMfyException
from llmfy.flow_engine.edge.edge import Edge
from llmfy.flow_engine.node.node import END, START, Node, NodeType
from llmfy.flow_engine.state.memory_manager import MemoryManager
from llmfy.flow_engine.state.workflow_state import WorkflowState
from llmfy.flow_engine.visualizer.visualizer import WorkflowVisualizer
from llmfy.llmfy_utils.deprecated.deprecated import deprecated


@deprecated(alternative='FlowEngine')
class LLMfyPipe:
    def __init__(
        self,
        initial_state: Optional[Dict[str, Any]] = None,
        memory: Optional[MemoryManager] = None,
    ):
        self.nodes: Dict[str, Node] = {
            START: Node(name=START, node_type=NodeType.START),
            END: Node(name=END, node_type=NodeType.END),
        }
        self.edges: List[Edge] = []
        self.state = WorkflowState(initial_state or {})
        self.visualizer = WorkflowVisualizer()
        self.memory = memory

    def validate_workflow(self) -> bool:
        """
        Validate the workflow structure.
        Ensures START has outgoing edges and END has incoming edges.
        """
        start_has_edge = False
        end_has_edge = False

        for edge in self.edges:
            if edge.source == START:
                start_has_edge = True
            if END in edge.targets:
                end_has_edge = True

        if not start_has_edge:
            raise ValueError("Workflow must have at least one edge from START")
        if not end_has_edge:
            raise ValueError("Workflow must have at least one edge to END")

        return True

    def add_node(
        self,
        name: str,
        func: Callable,
        node_type: NodeType = NodeType.FUNCTION,
        stream: bool = False,
    ) -> None:
        """
        Add a node to the workflow.
        Set is_streaming=True for nodes that return streaming content.

        Args:
            name (str): Node name
            func (Callable): Node description
            node_type (NodeType, optional): Node type. Defaults to NodeType.FUNCTION.
            stream (bool, optional): Node is use stream or not, if node use streaming set to True. Defaults to False.

        Raises:
            ValueError: _description_

        Returns:
            No return
        """
        if name in (START, END):
            raise ValueError(f"Cannot add node with reserved name {name}")

        # Modify the function to include state access
        original_func = func

        async def wrapped_func(*args, **kwargs):
            # Get the parameter signature of original_func
            sig = inspect.signature(original_func)
            if "state" in sig.parameters:
                kwargs["state"] = self.state

            # Check if original_func is an async function.
            if inspect.iscoroutinefunction(original_func):
                return await original_func(*args, **kwargs)
            return original_func(*args, **kwargs)

        # This makes the wrapper function appear like the original function
        wrapped_func.__signature__ = inspect.signature(original_func)  # type: ignore
        wrapped_func.__annotations__ = original_func.__annotations__

        sig = inspect.signature(original_func)
        sources = list(sig.parameters.keys())

        targets = []
        if sig.return_annotation != inspect.Signature.empty:
            if hasattr(sig.return_annotation, "__annotations__"):
                targets = list(sig.return_annotation.__annotations__.keys())

        node = Node(
            name=name,
            node_type=node_type,
            func=wrapped_func,
            sources=sources,
            targets=targets,
            stream=stream,
        )
        self.nodes[name] = node

    def add_edge(self, source: str, target: str) -> None:
        """
        Add a direct edge between nodes.

        Args:
            source (str): Node source name.
            target (str): Node target name.

        Raises:
            ValueError: _description_

        Returns:
            No return
        """
        if source not in self.nodes:
            raise ValueError(f"Source node '{source}' not found")
        if target not in self.nodes:
            raise ValueError(f"Target node '{target}' not found")

        edge = Edge(source=source, targets=[target])
        self.edges.append(edge)

    def add_conditional_edge(
        self,
        source: str,
        targets: Union[str, List[str]],
        condition: Callable,
    ) -> None:
        """
        Add a conditional edge with multiple possible targets.

        Args:
            source (str): Node source name.
            targets (Union[str, List[str]]): List node target name.
            condition (Callable): Conditional function.

        Raises:
            ValueError: _description_

        Returns:
            No return.
        """
        if source not in self.nodes:
            raise ValueError(f"Source node '{source}' not found")

        target_list = [targets] if isinstance(targets, str) else targets

        for target in target_list:
            if target not in self.nodes:
                raise ValueError(f"Target node '{target}' not found")

        original_condition = condition

        def wrapped_condition(*args, **kwargs):
            result = original_condition(self.state)
            if result not in target_list and result != END:
                raise ValueError(
                    f"Condition returned '{result}' which is not in targets: {target_list}"
                )
            return result

        edge = Edge(source=source, targets=target_list, condition=wrapped_condition)
        self.edges.append(edge)

    def _handle_state_update(self, additional_state, thread_id=None):
        updated_state = (
            self.memory.update_memory(thread_id, additional_state)
            if self.memory and thread_id
            else additional_state
        )
        self.state._update(updated_state)

    async def execute(
        self,
        additional_state: Dict[str, Any],
        thread_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Execute the workflow in standard (non-streaming) mode.

        Args:
            additional_state: Initial state to add to the workflow
            thread_id: Optional thread ID for memory management

        Returns:
            The final workflow state
        """
        try:
            # validate thread_id
            if self.memory and not thread_id:
                raise LLMfyException(
                    "`thread_id` required because workflow has memory."
                )

            if additional_state:
                self._handle_state_update(additional_state, thread_id)

            # Validate workflow before execution
            self.validate_workflow()

            nodes_to_process = [START]
            # processed_nodes = set()

            while nodes_to_process:
                current_node = nodes_to_process.pop(0)
                if current_node == END:
                    continue

                if current_node != START:
                    node = self.nodes[current_node]

                    try:
                        func_inputs = {
                            input_name: self.state.get(input_name)
                            for input_name in node.sources
                            if input_name != "state"
                            and input_name in self.state.get_current()
                        }

                        if node.func:
                            result = await node.func(**func_inputs)
                            # print(f"execute@LLMfyPipe result: {result}")
                            if isinstance(result, dict):
                                self._handle_state_update(result, thread_id)
                            elif len(node.targets) == 1:
                                self._handle_state_update(
                                    {node.targets[0]: result}, thread_id
                                )

                    except Exception as e:
                        raise LLMfyException(
                            f"Error executing node {current_node}: {str(e)}"
                        )

                # Get next nodes
                next_nodes = []
                for edge in self.edges:
                    if edge.source == current_node:
                        if edge.condition is None:
                            next_nodes.extend(edge.targets)
                        else:
                            target = edge.condition()
                            if target and target != END:
                                next_nodes.append(target)

                nodes_to_process.extend(next_nodes)

            return self.state.get_current()
        except Exception as e:
            raise LLMfyException(e)

    async def stream(
        self,
        additional_state: Dict[str, Any],
        thread_id: Optional[str] = None,
        stream_callback: Optional[Callable] = None,
    ) -> AsyncGenerator[Dict[str, Any], None]:
        """
        Execute the workflow in streaming mode, yielding results as they become available.

        Args:
            additional_state: Initial state to add to the workflow
            thread_id: Optional thread ID for memory management
            stream_callback: Optional callback function for handling streaming chunks (content only)

        Returns:
            An async generator yielding state and event updates
        """
        try:
            # validate thread_id
            if self.memory and not thread_id:
                raise LLMfyException(
                    "`thread_id` required because workflow has memory."
                )

            if additional_state:
                updated_state = self._handle_state_update(additional_state, thread_id)
                yield {"type": "state_update", "state": updated_state}

            # Validate workflow before execution
            self.validate_workflow()
            yield {"type": "workflow_start"}

            nodes_to_process = [START]

            while nodes_to_process:
                current_node = nodes_to_process.pop(0)

                if current_node == END:
                    yield {"type": "node_complete", "node": END}
                    continue

                yield {"type": "node_start", "node": current_node}

                if current_node != START:
                    node = self.nodes[current_node]

                    try:
                        func_inputs = {
                            input_name: self.state.get(input_name)
                            for input_name in node.sources
                            if input_name != "state"
                            and input_name in self.state.get_current()
                        }

                        if node.func:
                            if node.stream:
                                # Handle streaming node
                                async for chunk in await node.func(**func_inputs):
                                    if isinstance(chunk, dict):
                                        # Update state
                                        updated_state = self._handle_state_update(
                                            chunk, thread_id
                                        )
                                        yield {
                                            "type": "node_result",
                                            "node": current_node,
                                            "result": chunk,
                                            "state": updated_state,
                                        }

                                    else:
                                        # For raw non-dict chunks (like string tokens)
                                        yield {
                                            "type": "stream_chunk",
                                            "node": current_node,
                                            "content": chunk,
                                        }

                                        if stream_callback:
                                            await stream_callback(chunk)
                            else:
                                # Handle non-streaming node
                                result = await node.func(**func_inputs)
                                if isinstance(result, dict):
                                    updated_state = self._handle_state_update(
                                        result, thread_id
                                    )
                                    yield {
                                        "type": "node_result",
                                        "node": current_node,
                                        "result": result,
                                        "state": updated_state,
                                    }
                                elif len(node.targets) == 1:
                                    result_dict = {node.targets[0]: result}
                                    updated_state = self._handle_state_update(
                                        result_dict, thread_id
                                    )
                                    yield {
                                        "type": "node_result",
                                        "node": current_node,
                                        "result": result_dict,
                                        "state": updated_state,
                                    }

                    except Exception as e:
                        error_msg = f"Error executing node {current_node}: {str(e)}"
                        yield {
                            "type": "error",
                            "node": current_node,
                            "error": error_msg,
                        }
                        raise LLMfyException(error_msg)

                yield {"type": "node_complete", "node": current_node}

                # Get next nodes
                next_nodes = []
                for edge in self.edges:
                    if edge.source == current_node:
                        if edge.condition is None:
                            next_nodes.extend(edge.targets)
                        else:
                            target = edge.condition()
                            if target and target != END:
                                next_nodes.append(target)

                nodes_to_process.extend(next_nodes)
                yield {"type": "next_nodes", "nodes": next_nodes}

            yield {"type": "workflow_complete", "state": self.state.get_current()}
        except Exception as e:
            yield {"type": "workflow_error", "error": str(e)}
            raise LLMfyException(e)

    def get_diagram_code(self) -> str:
        """Get code for the workflow diagram."""
        return self.visualizer.create_mermaid_diagram(self)

    def get_diagram_base64(self):
        """Generate Mermaid diagram base64 for the workflow.

        Display with:
        ```py
        display(Image(url=your_diagram_url))
        ```
        """

        def get_base64(graph):
            graphbytes = graph.encode("utf8")
            base64_bytes = base64.urlsafe_b64encode(graphbytes)
            base64_string = base64_bytes.decode("ascii")
            return base64_string

        mermaid_code = self.get_diagram_code()
        return get_base64(
            f"""
			 {mermaid_code}
			 """
        )

    def get_diagram_url(self):
        """Generate Mermaid diagram url for the workflow.

        Display with:
        ```py
        display(Image(url=your_diagram_url))
        ```
        """

        def get_url(graph):
            graphbytes = graph.encode("utf8")
            base64_bytes = base64.urlsafe_b64encode(graphbytes)
            base64_string = base64_bytes.decode("ascii")
            return "https://mermaid.ink/img/" + base64_string

        mermaid_code = self.get_diagram_code()
        return get_url(
            f"""
			 {mermaid_code}
			 """
        )



from typing import Any

from pydantic import BaseModel


class Document(BaseModel):
    """Container for text document with dynamic metadata

    Example usage:
    ```python
    metadata = {
        "source": "llmfy",
    }
    meta = {
        "header_level": 1,
        "header_text": "Hello",
        **(metadata or {}),
    }
    doc = Document(
        id="doc_1",
        text="Sample",
        **meta,
    )
    print(doc)
    print(doc.source)
    print(doc.not_exist)
    ```
    """

    id: str
    text: str

    model_config = {
        "extra": "allow"  # Allow extra fields to be set dynamically (for metadata needs)
    }

    def __getattr__(self, name: str) -> Any:
        # Check normal attributes first
        try:
            return super().__getattribute__(name)
        except AttributeError:
            pass

        # Check Pydantic v2 dynamic fields
        extra = getattr(self, "model_extra", {})
        if name in extra:
            return extra[name]

        # Not found
        return None



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional


@dataclass
class CheckpointMetadata:
    """Metadata for a checkpoint."""
    checkpoint_id: str
    session_id: str
    timestamp: datetime
    node_name: str
    step: int



@dataclass
class Checkpoint:
    """Represents a saved state checkpoint."""
    metadata: CheckpointMetadata
    state: Dict[str, Any]
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert checkpoint to dictionary for storage."""
        return {
            "checkpoint_id": self.metadata.checkpoint_id,
            "session_id": self.metadata.session_id,
            "timestamp": self.metadata.timestamp.isoformat(),
            "node_name": self.metadata.node_name,
            "step": self.metadata.step,
            "state": self._serialize_state(self.state)
        }
    
    @staticmethod
    def _serialize_state(state: Dict[str, Any]) -> str:
        """Serialize state to JSON string, handling custom objects."""
        def default_serializer(obj):
            if hasattr(obj, '__dict__'):
                return {
                    '__type__': obj.__class__.__name__,
                    '__module__': obj.__class__.__module__,
                    'data': obj.__dict__
                }
            raise TypeError(f"Object of type {type(obj)} is not JSON serializable")
        
        return json.dumps(state, default=default_serializer)
    
    @staticmethod
    def _deserialize_state(state_str: str) -> Dict[str, Any]:
        """Deserialize state from JSON string, reconstructing custom objects."""
        import importlib
        
        def object_hook(dct):
            if '__type__' in dct and '__module__' in dct and 'data' in dct:
                # Reconstruct the custom object
                try:
                    module = importlib.import_module(dct['__module__'])
                    cls = getattr(module, dct['__type__'])
                    
                    # Handle different object construction patterns
                    if hasattr(cls, '__init__'):
                        # Try to create instance from data dict
                        obj_data = dct['data']
                        
                        # Check if class accepts **kwargs
                        try:
                            obj = cls(**obj_data)
                            return obj
                        except TypeError:
                            # If direct kwargs don't work, try creating empty instance
                            # and setting attributes
                            try:
                                obj = cls.__new__(cls)
                                for key, value in obj_data.items():
                                    setattr(obj, key, value)
                                return obj
                            except Exception as _:
                                # If all else fails, return the dict
                                return dct
                except (ImportError, AttributeError) as _:
                    # If we can't import the class, return the dict representation
                    return dct
            return dct
        
        return json.loads(state_str, object_hook=object_hook)
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Checkpoint":
        """Create checkpoint from dictionary."""
        metadata = CheckpointMetadata(
            checkpoint_id=data["checkpoint_id"],
            session_id=data["session_id"],
            timestamp=datetime.fromisoformat(data["timestamp"]),
            node_name=data["node_name"],
            step=data["step"]
        )
        state = cls._deserialize_state(data["state"])
        return cls(metadata=metadata, state=state)



class BaseCheckpointer(ABC):
    """Base class for checkpoint storage backends."""
    
    @abstractmethod
    async def save(self, checkpoint: Checkpoint) -> None:
        """
        Save a checkpoint.
        
        Args:
            checkpoint: The checkpoint to save
        """
        pass
    
    @abstractmethod
    async def load(self, session_id: str, checkpoint_id: Optional[str] = None) -> Optional[Checkpoint]:
        """
        Load a checkpoint.
        
        Args:
            session_id: The session ID
            checkpoint_id: Specific checkpoint ID, or None for latest
            
        Returns:
            The checkpoint if found, None otherwise
        """
        pass
    
    @abstractmethod
    async def list(self, session_id: str, limit: int = 10) -> List[Checkpoint]:
        """
        List checkpoints for a thread.
        
        Args:
            session_id: The session ID
            limit: Maximum number of checkpoints to return
            
        Returns:
            List of checkpoints, newest first
        """
        pass
    
    @abstractmethod
    async def delete(self, session_id: str, checkpoint_id: Optional[str] = None) -> None:
        """
        Delete checkpoint(s).
        
        Args:
            session_id: The session ID
            checkpoint_id: Specific checkpoint ID, or None to delete all for thread
        """
        pass
    
    @abstractmethod
    async def clear_all(self) -> None:
        """Clear all checkpoints from storage."""
        pass


import re


def extract_param_desc(param_name: str, docstring: str):
    """Extract parameter description from docstring.

    Args:
        param_name (str): Param name.
        docstring (str): Docstring.

    Returns:
        Parameter description.
    """
    param_description = ""

    # Google style: "param_name (type): description"
    google_match = re.search(
        rf"{re.escape(param_name)}\s*\([^)]*\)\s*:\s*(.+?)(?=\n\s*\w+\s*\([^)]*\)\s*:|\n\s*Returns?:|\n\s*Raises?:|\n\s*Examples?:|\n\s*Note:|\n\s*Yields?:|\n\s*Attributes?:|\n\s*\n|\Z)",
        docstring,
        re.DOTALL,
    )
    if google_match:
        # Clean up multi-line description
        param_description = " ".join(google_match.group(1).split())
    else:
        # reST style: ":param param_name: description"
        rest_match = re.search(
            rf":param\s+{re.escape(param_name)}\s*:\s*(.+?)(?=\n\s*:\w+|\n\s*\n|\Z)",
            docstring,
            re.DOTALL,
        )
        if rest_match:
            param_description = " ".join(rest_match.group(1).split())
        else:
            # Sphinx style with type: ":param type param_name: description"
            sphinx_match = re.search(
                rf":param\s+\w+\s+{re.escape(param_name)}\s*:\s*(.+?)(?=\n\s*:\w+|\n\s*\n|\Z)",
                docstring,
                re.DOTALL,
            )
            if sphinx_match:
                param_description = " ".join(sphinx_match.group(1).split())

    return param_description



import os
import re
import warnings
from typing import Any, Dict, List, Optional

from llmfy.exception.llmfy_exception import LLMfyException
from llmfy.llmfy_core.models.bedrock.bedrock_pricing_list import BEDROCK_PRICING
from llmfy.llmfy_core.models.model_pricing import ModelPricing
from llmfy.llmfy_core.models.openai.openai_pricing_list import OPENAI_PRICING
from llmfy.llmfy_core.service_provider import ServiceProvider
from llmfy.llmfy_core.service_type import ServiceType


class LLMfyUsage:
    """
    LLMfyUsage class.

    Count all provider token usage.

    Usage in `OpenAIModel`, `openai_usage_tracker` and `track_openai_usage`
    """

    def __init__(
        self,
        openai_pricing: Optional[Dict[str, Any]] = None,
        bedrock_pricing: Optional[Dict[str, Any]] = None,
    ):
        if openai_pricing:
            if not self.__is_valid_openai_pricing_structure(openai_pricing):
                error = """
				Please provide the right pricing structure for openai, example:
				```
				{
					"gpt-4o": {
						"input": 2.50,
						"output": 10.00
					},
					"gpt-4o-mini": {
						"input": 0.15,
						"output": 0.60
					},
					"gpt-3.5-turbo": {
						"input": 0.05,
						"output": 1.50
					}
				}
				```
				"""
                raise LLMfyException(error)

        if bedrock_pricing:
            if not self.__is_valid_bedrock_pricing_structure(bedrock_pricing):
                error = """
                Please provide the right pricing structure for bedrock, example:
                ```
                {
                    "anthropic.claude-3-5-sonnet-20240620-v1:0": {
                        "us-east-1": {
                            "region": "US East (N. Virginia)",
                            "input": 0.003,
                            "output": 0.015,
                        },
                        "us-west-2": {
                            "region": "US West (Oregon)",
                            "input": 0.003,
                            "output": 0.015,
                        },
                    },
                    "anthropic.claude-3-5-sonnet-20241022-v2:0": {
                        "us-east-1": {
                            "region": "US East (N. Virginia)",
                            "input": 0.003,
                            "output": 0.015,
                        },
                        "us-west-2": {
                            "region": "US West (Oregon)",
                            "input": 0.003,
                            "output": 0.015,
                        },
                    }
                }
                ```
                """
                raise LLMfyException(error)

        self.total_request: int = 0
        self.output_tokens: int = 0
        self.input_tokens: int = 0
        self.total_tokens: int = 0
        self.total_cost: float = 0.0
        self.raw_usages: List[Dict[str, int]] = []
        self.openai_pricing: Dict[str, ModelPricing] = (
            self._load_openai_pricing(pricing_source=openai_pricing or OPENAI_PRICING)
            or {}
        )
        self.bedrock_pricing: Dict[str, Dict[str, ModelPricing]] = (
            self._load_bedrock_pricing(
                pricing_source=bedrock_pricing or BEDROCK_PRICING
            )
            or {}
        )
        self.details: List[Dict[str, Any]] = []

    def to_dict(self):
        return {
            "total_request": self.total_request,
            "tokens": {
                "total_tokens": self.total_tokens,
                "input_tokens": self.input_tokens,
                "output_tokens": self.output_tokens,
            },
            "costs": {
                "total_cost": self.total_cost,
                "total_cost_formatted": self.__format_trimmed_float(self.total_cost),
            },
            "details": self.details,
        }

    def __repr__(self) -> str:
        # rounded_up = math.ceil(self.total_cost * 1000000) / 1000000
        return (
            f"\n------------------\n"
            f"USAGE: \n\n"
            f"Total Tokens: {self.total_tokens}\n"
            f"\tInput Tokens: {self.input_tokens}\n"
            f"\tOutput Tokens: {self.output_tokens}\n"
            f"Total Requests: {self.total_request}\n"
            # f"Total Cost (USD): ${rounded_up:.6f}"
            f"Total Cost (USD): ${self.total_cost}\n"
            f"Total Cost (USD formatted): ${self.__format_trimmed_float(self.total_cost)}\n\n"
            f"Request Details:\n"
            + "\n".join(
                f"{i + 1}. {detail['model']} "
                f"\n\tprovider: {detail['provider']} "
                f"\n\ttype: {detail['type']} "
                f"\n\tinput_tokens: {detail['input_tokens']} "
                f"\n\toutput_tokens: {detail['output_tokens']} "
                f"\n\ttotal_tokens: {detail['total_tokens']} "
                f"\n\tinput_price: {detail['input_price']} "
                f"\n\toutput_price: {detail['output_price']} "
                f"\n\tprice_per_tokens: {detail['price_per_tokens']} "
                f"\n\ttotal_cost (USD): {detail['total_cost']} "
                f"\n\ttotal_cost (USD formatted): {self.__format_trimmed_float(detail.get('total_cost', 0))}\n"
                for i, detail in enumerate(self.details)
            )
            + "\n------------------\n"
        )

    def __format_trimmed_float(self, value: float) -> str:
        """
        Format a float with up to 9 decimal places,
        trimming unnecessary trailing zeros.
        """
        formatted = f"{value:.9f}"
        if "." in formatted:
            formatted = formatted.rstrip("0").rstrip(".")
        return formatted

    def __is_valid_openai_pricing_structure(self, pricing: Any) -> bool:
        if not isinstance(pricing, dict):
            return False
        for outer_dict in pricing.values():
            if not isinstance(outer_dict, dict):
                return False
            for value in outer_dict.values():
                if not isinstance(value, (float, int)):
                    return False
        return True

    def __is_valid_bedrock_pricing_structure(self, pricing: Any) -> bool:
        # First check if it's a dictionary
        if not isinstance(pricing, dict):
            return False

        # Check first level values are dictionaries
        for outer_dict in pricing.values():
            if not isinstance(outer_dict, dict):
                return False

            # Check second level values are dictionaries
            for inner_dict in outer_dict.values():
                if not isinstance(inner_dict, dict):
                    return False

        return True

    def _load_openai_pricing(
        self,
        pricing_source: Dict[str, Dict[str, Any]],
    ) -> Dict[str, ModelPricing]:
        """
        Load openai pricing from dictionary.

        Price per 1M tokens for different models (USD)
        - https://platform.openai.com/docs/pricing
        """
        pricing_data = pricing_source
        return {
            model: ModelPricing(token_input=data["input"], token_output=data["output"])
            for model, data in pricing_data.items()
        }

    def _load_bedrock_pricing(
        self,
        pricing_source: Dict[str, Dict[str, Dict[str, Any]]],
    ) -> Dict[str, Dict[str, ModelPricing]]:
        """
        Load bedrock pricing from dictionary.

        Price per 1K tokens for different models (USD)
        - https://aws.amazon.com/bedrock/pricing/
        - https://aws.amazon.com/bedrock/pricing/
        - https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/Concepts.RegionsAndAvailabilityZones.html
        """
        pricing_data = pricing_source
        return {
            model: {
                region: ModelPricing(
                    token_input=pricing["input"],
                    token_output=pricing["output"],
                )
                for region, pricing in regions.items()
            }
            for model, regions in pricing_data.items()
        }

    def update(
        self,
        provider: ServiceProvider,
        type: ServiceType,
        model: str,
        usage: Dict[str, int],
    ) -> None:
        """
        Update usage statistics and calculate price.

        Args:
            provider (ModelProvider): Model provider
            model (str): Model name
            usage (Dict[str, int]): Dictionary containing token counts
        """
        match type:
            case ServiceType.LLM:
                match provider:
                    case ServiceProvider.OPENAI:
                        self.__openai_update(model=model, usage=usage)
                    case ServiceProvider.BEDROCK:
                        self.__bedrock_update(model=model, usage=usage)
                    case _:
                        pass
            case ServiceType.EMBEDDING:
                match provider:
                    case ServiceProvider.OPENAI:
                        self.__openai_embedding_update(model=model, usage=usage)
                    case ServiceProvider.BEDROCK:
                        self.__bedrock_embedding_update(model=model, usage=usage)
                    case _:
                        pass
            case _:
                pass

    def __openai_update(
        self,
        model: str,
        usage: Dict[str, int],
    ) -> None:
        """
        OPENAI Update usage statistics and calculate price.

        Args:
            model (str): Model name
            usage (Dict[str, int]): Dictionary containing token counts
        """
        ONE_MILLION = 1000000

        self.raw_usages.append(usage)
        usage_dict = vars(usage) if hasattr(usage, "__dict__") else usage

        # usage per-request
        input_tokens = usage_dict.get("prompt_tokens", 0)
        output_tokens = usage_dict.get("completion_tokens", 0)
        total_tokens = input_tokens + output_tokens

        # usage accumulation
        self.total_request += 1
        self.input_tokens += input_tokens
        self.output_tokens += output_tokens
        self.total_tokens += total_tokens

        # Calculate price
        price_info = None
        total_cost_per_request = 0

        if model in self.openai_pricing:
            price_info = self.openai_pricing[model]
            # pricing per-request
            i_price = (input_tokens / ONE_MILLION) * price_info.token_input
            o_price = (output_tokens / ONE_MILLION) * price_info.token_output
            total_cost_per_request = i_price + o_price

            # pricing accumulation
            self.total_cost += total_cost_per_request
        else:
            warnings.warn(
                "MODEL not found at specified openai pricing. You can add in custom prices with `llmfy_usage_tracker(openai_pricing=prices)`"
            )

        # add to details per-requestd
        self.details.append(
            {
                "model": model,
                "provider": ServiceProvider.OPENAI,
                "type": ServiceType.LLM,
                "input_tokens": input_tokens,
                "output_tokens": output_tokens,
                "total_tokens": total_tokens,
                "input_price": price_info.token_input if price_info else None,
                "output_price": price_info.token_output if price_info else None,
                "price_per_tokens": ONE_MILLION,
                "total_cost": total_cost_per_request,
            }
        )
        pass

    def __openai_embedding_update(self, model: str, usage: Dict) -> None:
        """
        Update usage statistics and calculate price.

        Args:
            model: Model name
            usage: usage of input token (for embedding use input token only).
        """

        def has_cross_region_inference_id(s):
            return bool(re.match(r"^.{2}\.", s))

        ONE_MILLION = 1000000

        self.raw_usages.append(usage)
        usage_dict = vars(usage) if hasattr(usage, "__dict__") else usage

        # usage per-request
        input_tokens = usage_dict.get("prompt_tokens", 0)
        output_tokens = 0  # in embedding no output tokens usage
        total_tokens = input_tokens + output_tokens

        # usage accumulation
        self.total_request += 1
        self.input_tokens += input_tokens
        self.output_tokens += output_tokens
        self.total_tokens += total_tokens

        # Calculate price
        price_info = None
        total_cost_per_request = 0

        if model in self.openai_pricing:
            price_info = self.openai_pricing[model]
            # pricing per-request
            i_price = (input_tokens / ONE_MILLION) * price_info.token_input
            o_price = (output_tokens / ONE_MILLION) * price_info.token_output
            total_cost_per_request = i_price + o_price

            # pricing accumulation
            self.total_cost += total_cost_per_request
        else:
            warnings.warn(
                "MODEL not found at specified bedrock pricing. You can add in custom prices with `llmfy_usage_tracker(bedrock_pricing=prices)`"
            )

        # add to details per-request
        self.details.append(
            {
                "model": model,
                "provider": ServiceProvider.OPENAI,
                "type": ServiceType.EMBEDDING,
                "input_tokens": input_tokens,
                "output_tokens": output_tokens,
                "total_tokens": total_tokens,
                "input_price": price_info.token_input if price_info else None,
                "output_price": price_info.token_output if price_info else None,
                "price_per_tokens": ONE_MILLION,
                "total_cost": total_cost_per_request,
            }
        )
        pass

    def __bedrock_update(self, model: str, usage: Dict[str, int]) -> None:
        """
        Update usage statistics and calculate price.

        Args:
            model: Model name
            usage: Dictionary containing token counts
        """

        def has_cross_region_inference_id(s):
            return bool(re.match(r"^.{2}\.", s))

        ONE_K = 1000
        AWS_REGION = os.getenv("AWS_BEDROCK_REGION") or ""
        MODEL = model[3:] if has_cross_region_inference_id(model) else model

        self.raw_usages.append(usage)
        usage_dict = vars(usage) if hasattr(usage, "__dict__") else usage

        # usage per-request
        input_tokens = usage_dict.get("inputTokens", 0)
        output_tokens = usage_dict.get("outputTokens", 0)
        total_tokens = input_tokens + output_tokens

        # usage accumulation
        self.total_request += 1
        self.input_tokens += input_tokens
        self.output_tokens += output_tokens
        self.total_tokens += total_tokens

        # Calculate price
        price_info = None
        total_cost_per_request = 0

        if MODEL in self.bedrock_pricing:
            price_info = self.bedrock_pricing[MODEL][AWS_REGION]

            # pricing per-request
            i_price = (input_tokens / ONE_K) * price_info.token_input
            o_price = (output_tokens / ONE_K) * price_info.token_output
            total_cost_per_request = i_price + o_price

            # pricing accumulation
            self.total_cost += total_cost_per_request
        else:
            warnings.warn(
                "MODEL not found at specified bedrock pricing. You can add in custom prices with `llmfy_usage_tracker(bedrock_pricing=prices)`"
            )

        # add to details per-request
        self.details.append(
            {
                "model": model,
                "provider": ServiceProvider.BEDROCK,
                "type": ServiceType.LLM,
                "input_tokens": input_tokens,
                "output_tokens": output_tokens,
                "total_tokens": total_tokens,
                "input_price": price_info.token_input if price_info else None,
                "output_price": price_info.token_output if price_info else None,
                "price_per_tokens": ONE_K,
                "total_cost": total_cost_per_request,
            }
        )
        pass

    def __bedrock_embedding_update(self, model: str, usage: Dict) -> None:
        """
        Update usage statistics and calculate price.

        Args:
            model: Model name
            usage: usage of input token (for embedding use input token only).
        """

        def has_cross_region_inference_id(s):
            return bool(re.match(r"^.{2}\.", s))

        ONE_K = 1000
        AWS_REGION = os.getenv("AWS_BEDROCK_REGION") or ""
        MODEL = model[3:] if has_cross_region_inference_id(model) else model

        self.raw_usages.append(usage)
        usage_dict = vars(usage) if hasattr(usage, "__dict__") else usage

        # usage per-request
        input_tokens = usage_dict.get("x-amzn-bedrock-input-token-count", 0)
        output_tokens = 0  # in embedding no output tokens usage
        total_tokens = input_tokens + output_tokens

        # usage accumulation
        self.total_request += 1
        self.input_tokens += input_tokens
        self.output_tokens += output_tokens
        self.total_tokens += total_tokens

        # Calculate price
        price_info = None
        total_cost_per_request = 0

        if MODEL in self.bedrock_pricing:
            price_info = self.bedrock_pricing[MODEL][AWS_REGION]

            # pricing per-request
            i_price = (input_tokens / ONE_K) * price_info.token_input
            o_price = 0  # in embedding no output tokens usage
            total_cost_per_request = i_price + o_price

            # pricing accumulation
            self.total_cost += total_cost_per_request
        else:
            warnings.warn(
                "MODEL not found at specified bedrock pricing. You can add in custom prices with `llmfy_usage_tracker(bedrock_pricing=prices)`"
            )

        # add to details per-request
        self.details.append(
            {
                "model": model,
                "provider": ServiceProvider.BEDROCK,
                "type": ServiceType.EMBEDDING,
                "input_tokens": input_tokens,
                "output_tokens": output_tokens,
                "total_tokens": total_tokens,
                "input_price": price_info.token_input if price_info else None,
                "output_price": price_info.token_output if price_info else None,
                "price_per_tokens": ONE_K,
                "total_cost": total_cost_per_request,
            }
        )
        pass

    def reset(self):
        """
        Reset usage statistics and calculate price.
        """
        self.total_request = 0
        self.input_tokens = 0
        self.output_tokens = 0
        self.total_tokens = 0
        self.total_cost = 0.0
        self.raw_usages = []
        self.details = []



"""
LLMfy Exception Handling

Documentation References:

- AWS Bedrock Converse API: [https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Converse.html](https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Converse.html)
- OpenAI Python SDK: [https://github.com/openai/openai-python#handling-errors](https://github.com/openai/openai-python#handling-errors)
- Google Gen AI SDK: [https://github.com/googleapis/python-genai#error-handling](https://github.com/googleapis/python-genai#error-handling)

Example of catching and inspecting an error:
```python
try:
    # _call_llmfy
    pass
except LLMfyException as e:
    print(f"Error: {e.message}")
    print(f"Status Code: {e.status_code}")
    print(f"Provider: {e.provider}")
    print(f"Raw Error: {e.raw_error}")

    # Check specific exception type
    if isinstance(e, RateLimitException):
        print("Rate limited! Implement backoff...")
    elif isinstance(e, TimeoutException):
        print("Request timed out! Retry...")
```
"""

from typing import Any, Optional


class LLMfyException(Exception):
    """
    Base LLMfy Exception

    Example of catching and inspecting an error:
    ```python
    try:
        # _call_llmfy
        pass
    except LLMfyException as e:
        print(f"Error: {e.message}")
        print(f"Status Code: {e.status_code}")
        print(f"Provider: {e.provider}")
        print(f"Raw Error: {e.raw_error}")

        # Check specific exception type
        if isinstance(e, RateLimitException):
            print("Rate limited! Implement backoff...")
        elif isinstance(e, TimeoutException):
            print("Request timed out! Retry...")
    ```
    """

    def __init__(
        self,
        message: str,
        status_code: Optional[int] = None,
        raw_error: Optional[Any] = None,
        provider: Optional[str] = None,
    ):
        super().__init__(message)
        self.message = message
        self.status_code = status_code
        self.raw_error = raw_error
        self.provider = provider

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"message={self.message!r}, "
            f"status_code={self.status_code}, "
            f"provider={self.provider!r})"
        )


class RateLimitException(LLMfyException):
    """Rate limit exceeded"""

    pass


class QuotaExceededException(LLMfyException):
    """Quota/usage limit exceeded"""

    pass


class TimeoutException(LLMfyException):
    """Request timed out"""

    pass


class InvalidRequestException(LLMfyException):
    """Invalid request parameters"""

    pass


class AuthenticationException(LLMfyException):
    """Authentication failed"""

    pass


class PermissionDeniedException(LLMfyException):
    """Permission denied"""

    pass


class ModelNotFoundException(LLMfyException):
    """Model not found or unavailable"""

    pass


class ServiceUnavailableException(LLMfyException):
    """Service temporarily unavailable"""

    pass


class ContentFilterException(LLMfyException):
    """Content blocked by safety filters"""

    pass


class ModelErrorException(LLMfyException):
    """Model processing error"""

    pass



from ._version import __version__

from .exception import (
    AuthenticationException,
    ContentFilterException,
    InvalidRequestException,
    LLMfyException,
    ModelErrorException,
    ModelNotFoundException,
    PermissionDeniedException,
    QuotaExceededException,
    RateLimitException,
    ServiceUnavailableException,
    TimeoutException,
)
from .flow_engine import (
    END,
    START,
    BaseCheckpointer,
    Edge,
    FlowEngine,
    FlowEngineStreamResponse,
    FlowEngineStreamType,
    InMemoryCheckpointer,
    MemoryManager,
    Node,
    NodeStreamResponse,
    NodeStreamType,
    RedisCheckpointer,
    SQLCheckpointer,
    ToolNodeStreamResponse,
    ToolNodeStreamType,
    WorkflowState,
    count_tokens_approximately,
    safe_trim_messages,
    tools_node,
    tools_stream_node,
    trim_messages,
)
from .llmfy_core import (
    AIResponse,
    BedrockConfig,
    BedrockModel,
    Content,
    ContentType,
    GenerationResponse,
    LLMfy,
    LLMfyUsage,
    Message,
    OpenAIConfig,
    OpenAIModel,
    Role,
    Tool,
    ToolRegistry,
    llmfy_usage_tracker,
)
from .llmfy_core.embeddings.base_embedding_model import BaseEmbeddingModel
from .llmfy_core.embeddings.bedrock.bedrock_embedding import BedrockEmbedding
from .llmfy_core.embeddings.openai.openai_embedding import OpenAIEmbedding
from .llmfy_utils.chunk import chunk_markdown_by_header, chunk_text
from .llmfy_utils.text_preprocessing import clean_text_for_embedding
from .vector_store.document import Document
from .vector_store.faiss_index.faiss_index import FAISSIndex
from .vector_store.faiss_index.faiss_vector_store import FAISSVectorStore

__all__ = [
    "LLMfy",
    "Message",
    "Role",
    "Tool",
    "ToolRegistry",
    "AIResponse",
    "GenerationResponse",
    "OpenAIConfig",
    "OpenAIModel",
    "BedrockConfig",
    "BedrockModel",
    "LLMfyUsage",
    "llmfy_usage_tracker",
    "Content",
    "ContentType",
    "Document",
    "FAISSIndex",
    "FAISSVectorStore",
    "BaseEmbeddingModel",
    "BedrockEmbedding",
    "OpenAIEmbedding",
    "chunk_text",
    "clean_text_for_embedding",
    "chunk_markdown_by_header",
    "FlowEngine",
    "Edge",
    "Node",
    "START",
    "END",
    "WorkflowState",
    "MemoryManager",
    "BaseCheckpointer",
    "InMemoryCheckpointer",
    "RedisCheckpointer",
    "SQLCheckpointer",
    "tools_node",
    "tools_stream_node",
    "trim_messages",
    "safe_trim_messages",
    "count_tokens_approximately",
    "FlowEngineStreamResponse",
    "FlowEngineStreamType",
    "NodeStreamResponse",
    "NodeStreamType",
    "ToolNodeStreamResponse",
    "ToolNodeStreamType",
    "LLMfyException",
    "AuthenticationException",
    "ContentFilterException",
    "InvalidRequestException",
    "ModelErrorException",
    "ModelNotFoundException",
    "PermissionDeniedException",
    "QuotaExceededException",
    "RateLimitException",
    "ServiceUnavailableException",
    "TimeoutException",
    "__version__",
]



import json
from typing import Optional

from llmfy.exception.llmfy_exception import LLMfyException
from llmfy.flow_engine.checkpointer.base_checkpointer import (
    BaseCheckpointer,
    Checkpoint,
)

try:
    import redis.asyncio as redis

    REDIS_AVAILABLE = True
except ImportError:
    REDIS_AVAILABLE = False


class RedisCheckpointer(BaseCheckpointer):
    """Redis checkpoint storage backend."""

    def __init__(
        self,
        redis_url: str = "redis://localhost:6379/0",
        prefix: str = "llmfy_checkpoint:",
        ttl: Optional[int] = None,
    ):
        """
        Initialize the Redis checkpointer.

        Args:
            redis_url: Redis connection URL
            prefix: Key prefix for checkpoints
            ttl: Time-to-live in seconds for checkpoints (None = no expiration)
        """
        if not REDIS_AVAILABLE:
            raise LLMfyException(
                "redis package is not installed. redis package is required for RedisCheckpointer. "
                'Install it using `pip install "llmfy[redis]"`'
            )

        self.redis_url = redis_url
        self.prefix = prefix
        self.ttl = ttl
        self._client: Optional[redis.Redis] = None

    async def _get_client(self) -> redis.Redis:
        """Get or create Redis client."""
        if self._client is None:
            self._client = await redis.from_url(
                self.redis_url, encoding="utf-8", decode_responses=True
            )
        return self._client

    def _session_key(self, session_id: str) -> str:
        """Get Redis key for session's checkpoint list."""
        return f"{self.prefix}session:{session_id}"

    def _checkpoint_key(self, checkpoint_id: str) -> str:
        """Get Redis key for specific checkpoint."""
        return f"{self.prefix}checkpoint:{checkpoint_id}"

    async def save(self, checkpoint: Checkpoint) -> None:
        """
        Save a checkpoint to Redis.

        Args:
            checkpoint: The checkpoint to save
        """
        client = await self._get_client()

        session_id = checkpoint.metadata.session_id
        checkpoint_id = checkpoint.metadata.checkpoint_id
        timestamp = checkpoint.metadata.timestamp.timestamp()

        # Save checkpoint data
        checkpoint_key = self._checkpoint_key(checkpoint_id)
        checkpoint_data = json.dumps(checkpoint.to_dict())

        await client.set(checkpoint_key, checkpoint_data)

        if self.ttl:
            await client.expire(checkpoint_key, self.ttl)

        # Add to thread's sorted set (sorted by timestamp)
        session_key = self._session_key(session_id)
        await client.zadd(session_key, {checkpoint_id: timestamp})

        if self.ttl:
            await client.expire(session_key, self.ttl)

    async def load(
        self,
        session_id: str,
        checkpoint_id: Optional[str] = None,
    ) -> Optional[Checkpoint]:
        """
        Load a checkpoint from Redis.

        Args:
            session_id: The session ID
            checkpoint_id: Specific checkpoint ID, or None for latest

        Returns:
            The checkpoint if found, None otherwise
        """
        client = await self._get_client()

        if checkpoint_id is None:
            # Get latest checkpoint ID from sorted set
            session_key = self._session_key(session_id)
            results = await client.zrevrange(session_key, 0, 0)

            if not results:
                return None

            checkpoint_id = results[0]

        # Load checkpoint data
        checkpoint_key = self._checkpoint_key(checkpoint_id)  # type: ignore
        data = await client.get(checkpoint_key)

        if data is None:
            return None

        checkpoint_dict = json.loads(data)
        return Checkpoint.from_dict(checkpoint_dict)

    async def list(self, session_id: str, limit: int = 10) -> list[Checkpoint]:
        """
        List checkpoints for a thread.

        Args:
            session_id: The session ID
            limit: Maximum number of checkpoints to return

        Returns:
            List of checkpoints, newest first
        """
        client = await self._get_client()

        # Get checkpoint IDs from sorted set (newest first)
        thread_key = self._session_key(session_id)
        checkpoint_ids = await client.zrevrange(thread_key, 0, limit - 1)

        if not checkpoint_ids:
            return []

        # Load all checkpoints
        checkpoints = []
        for checkpoint_id in checkpoint_ids:
            checkpoint = await self.load(session_id, checkpoint_id)
            if checkpoint:
                checkpoints.append(checkpoint)

        return checkpoints

    async def delete(self, session_id: str, checkpoint_id: Optional[str] = None) -> None:
        """
        Delete checkpoint(s) from Redis.

        Args:
            session_id: The session ID
            checkpoint_id: Specific checkpoint ID, or None to delete all for thread
        """
        client = await self._get_client()
        session_key = self._session_key(session_id)

        if checkpoint_id:
            # Delete specific checkpoint
            checkpoint_key = self._checkpoint_key(checkpoint_id)
            await client.delete(checkpoint_key)
            await client.zrem(session_key, checkpoint_id)
        else:
            # Delete all checkpoints for thread
            checkpoint_ids = await client.zrange(session_key, 0, -1)

            if checkpoint_ids:
                # Delete all checkpoint data
                checkpoint_keys = [self._checkpoint_key(cid) for cid in checkpoint_ids]
                await client.delete(*checkpoint_keys)

            # Delete session sorted set
            await client.delete(session_key)

    async def clear_all(self) -> None:
        """Clear all checkpoints from Redis."""
        client = await self._get_client()

        # Find all keys with our prefix
        pattern = f"{self.prefix}*"
        cursor = 0

        while True:
            cursor, keys = await client.scan(cursor, match=pattern, count=100)

            if keys:
                await client.delete(*keys)

            if cursor == 0:
                break

    async def close(self) -> None:
        """Close the Redis connection."""
        if self._client:
            await self._client.close()
            self._client = None

    async def __aenter__(self):
        """Async context manager entry."""
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        await self.close()



from copy import deepcopy
from typing import Any, Generator, List

from llmfy.flow_engine.stream.tool_node_stream_response import (
    ToolNodeStreamResponse,
    ToolNodeStreamType,
)
from llmfy.llmfy_core.messages.message import Message
from llmfy.llmfy_core.messages.role import Role
from llmfy.llmfy_core.tools.tool_registry import ToolRegistry


def tools_node(messages: List[Message], registry: ToolRegistry) -> List[Message]:
    """Tools Node

    Args:
        messages (List[Message]): List message
        registry (ToolRegistry): Tool registry

    Returns:
        Tool results message
    """
    new_messages = deepcopy(messages)
    last_message = new_messages[-1]
    tool_results: List[Message] = []
    if last_message.tool_calls:
        for tool_call in last_message.tool_calls:
            result = registry.execute_tool(
                name=tool_call.name, arguments=tool_call.arguments
            )
            tool_results.append(
                Message(
                    role=Role.TOOL,
                    request_call_id=tool_call.request_call_id,
                    tool_call_id=tool_call.tool_call_id,
                    tool_results=[str(result)],
                    name=tool_call.name,
                )
            )
    return tool_results


def tools_stream_node(messages: List[Message], registry: ToolRegistry) -> Generator[ToolNodeStreamResponse, Any, None]:
    """Tools Stream Node

    Args:
        messages (List[Message]): List message
        registry (ToolRegistry): Tool registry

    Yields:
        Tool results message
    """
    new_messages = deepcopy(messages)
    last_message = new_messages[-1]
    response = ToolNodeStreamResponse()

    if last_message.tool_calls:
        for tool_call in last_message.tool_calls:
            # Executing
            response.type = ToolNodeStreamType.EXECUTING
            response.name = tool_call.name
            response.arguments = tool_call.arguments
            response.result = None
            yield response

            result = registry.execute_tool(
                name=tool_call.name, arguments=tool_call.arguments
            )
            tool_result = Message(
                role=Role.TOOL,
                request_call_id=tool_call.request_call_id,
                tool_call_id=tool_call.tool_call_id,
                tool_results=[str(result)],
                name=tool_call.name,
            )

            # Result
            response.type = ToolNodeStreamType.RESULT
            response.name = tool_call.name
            response.arguments = tool_call.arguments
            response.result = tool_result
            yield response



from .flow_engine_stream_response import FlowEngineStreamResponse, FlowEngineStreamType
from .node_stream_response import NodeStreamResponse, NodeStreamType
from .tool_node_stream_response import ToolNodeStreamResponse, ToolNodeStreamType

__all__ = [
    "FlowEngineStreamResponse",
    "FlowEngineStreamType",
    "NodeStreamResponse",
    "NodeStreamType",
    "ToolNodeStreamResponse",
    "ToolNodeStreamType",
]



from contextlib import contextmanager
from contextvars import ContextVar
from typing import Any, Dict, Optional

from llmfy.llmfy_core.usage.llmfy_usage import LLMfyUsage

# Thread-safe storage for token usage per request
LLMFY_USAGE_TRACKER_VAR = ContextVar("LLMFY_USAGE_TRACKER", default=LLMfyUsage())


@contextmanager
def llmfy_usage_tracker(
    openai_pricing: Optional[Dict[str, Any]] = None,
    bedrock_pricing: Optional[Dict[str, Any]] = None,
):
    """LLMfy usage tracker.

    Use this to track token usage all provider.

    Example:
    ```python
    with llmfy_usage_tracker() as usage:
            result = llm.generate(messages)
            ...
            print(usage)
    ```

    [OPENAI] Price per 1M tokens for different models (USD):
        - https://platform.openai.com/docs/pricing

    [BEDROCK] Price per 1K tokens for different models (USD):
        - https://aws.amazon.com/bedrock/pricing/
        - https://aws.amazon.com/bedrock/pricing/
        - https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/Concepts.RegionsAndAvailabilityZones.html

    Args:
        openai_pricing (Optional[Dict[str, Any]], optional): OpenAI Pricing dictionary source. Defaults to None.
            If None then use default pricing from this dependency.

            Example pricing structure:
            ```
            {
                "gpt-4o": {
                    "input": 2.50,
                    "output": 10.00
                },
                "gpt-4o-mini": {
                    "input": 0.15,
                    "output": 0.60
                },
                "gpt-3.5-turbo": {
                    "input": 0.05,
                    "output": 1.50
                }
            }
            ```

        bedrock_pricing (Optional[Dict[str, Any]], optional): Bedrock Pricing dictionary source. Defaults to None.
            If None then use default pricing from this dependency.

            Example pricing structure:
            ```
            {
                "anthropic.claude-3-5-sonnet-20240620-v1:0": {
                    "us-east-1": {
                        "region": "US East (N. Virginia)",
                        "input": 0.003,
                        "output": 0.015,
                    },
                    "us-west-2": {
                        "region": "US West (Oregon)",
                        "input": 0.003,
                        "output": 0.015,
                    },
                },
                "anthropic.claude-3-5-sonnet-20241022-v2:0": {
                    "us-east-1": {
                        "region": "US East (N. Virginia)",
                        "input": 0.003,
                        "output": 0.015,
                    },
                    "us-west-2": {
                        "region": "US West (Oregon)",
                        "input": 0.003,
                        "output": 0.015,
                    },
                }
            }
            ```

    Yields:
            OpenAIUsage: OpenAI usage accumulation.
    """

    usage_tracker = LLMfyUsage(
        openai_pricing=openai_pricing,
        bedrock_pricing=bedrock_pricing,
    )
    LLMFY_USAGE_TRACKER_VAR.set(usage_tracker)  # Store usage_tracker it in the context
    try:
        yield usage_tracker  # Expose tracker to the context
    finally:
        pass



import uuid
from datetime import datetime, timezone
from typing import Any, List, Optional

from pydantic import BaseModel, ConfigDict, Field

from llmfy.llmfy_core.messages.content import Content
from llmfy.llmfy_core.messages.role import Role
from llmfy.llmfy_core.messages.tool_call import ToolCall


class Message(BaseModel):
    """Message class for input to the LLM models.

    Args:
        id (str): Id message default UUIDv4.
        role (Role): Message role.
        content (Optional[str] | Optional[List[Content]]): Use str if only using text, but if use image and text use List[Content].
        name (Optional[str]): Message name.
        tool_calls (Optional[List[ToolCall]]): [`assistant` role ONLY] Tool call list.
        tool_call_id (Optional[str]): [`tool` role ONLY] Tool call id.
        tool_results (Optional[List[Any]]): [`tool` role ONLY] Tool call results.
        request_call_id (Optional[str]): [`tool` role ONLY] Tool call id request.
    """

    model_config = ConfigDict(extra="forbid")

    id: str = Field(default_factory=lambda: str(uuid.uuid4()))

    role: Role
    """Message role [SYSTEM, USER, ASSISTANT, TOOL]"""

    content: Optional[str] | Optional[List[Content]] = None
    """Use str if only using text, but if use image and text use List[Content]."""

    name: Optional[str] = None
    """Message name"""

    tool_calls: Optional[List[ToolCall]] = None  # For Message with `assistant` role
    """[`assistant` role ONLY] Tool call list."""

    tool_call_id: Optional[str] = None  # For Message with `tool` role
    """[`tool` role ONLY] Tool call id."""

    tool_results: Optional[List[Any]] = None  # For Message with `tool` role
    """[`tool` role ONLY] Tool call results."""

    request_call_id: Optional[str] = None  # For Message with `tool` role
    """[`tool` role ONLY] Tool call id request."""

    timestamp: str = Field(
        default_factory=lambda: datetime.now(timezone.utc).isoformat(),
        description="Message timestamp in ISO 8601. Default UTC",
    )
    """Message timestamp in ISO 8601. Default UTC"""

    # def __init__(self, **kwargs):
    def model_post_init(self, __context) -> None:
        # Ensure tool_results is only used when role is "tool"
        if self.tool_results is not None and self.role != Role.TOOL:
            raise ValueError("tool_results can only be set when role is 'tool'.")

        # Ensure expect tool_results used when role is "tool"
        if not self.tool_results and self.role == Role.TOOL:
            raise ValueError("Expected tool_results when role is 'tool'.")

        # Ensure tool_call_id is only used when role is "tool"
        if self.tool_call_id is not None and self.role != Role.TOOL:
            raise ValueError("tool_call_id can only be set when role is 'tool'.")

        # Ensure tool_calls is only used when role is "assistant"
        if self.tool_calls is not None and self.role != Role.ASSISTANT:
            raise ValueError("tool_calls can only be set when role is 'assistant'.")



from enum import Enum


class ContentType(str, Enum):
    """ContentType enum for `Content` class."""

    TEXT = "text"
    IMAGE = "image"
    DOCUMENT = "document"
    VIDEO = "video"

    def __str__(self):
        return self.value

    def __repr__(self):
        return f"'{self.value}'"



from enum import Enum
from typing import Any, Dict, Optional

from pydantic import BaseModel, Field


class ToolNodeStreamType(str, Enum):
    """ToolNodeStreamType"""
    EXECUTING = "executing"
    RESULT = "result"

class ToolNodeStreamResponse(BaseModel):
    """ToolNodeStreamResponse"""
    type: Optional[str] = Field(default=None)
    name: Optional[str] = Field(default=None)
    arguments: Optional[Dict] = Field(default=None)
    result: Optional[Any] = Field(default=None)


from typing import Any, Literal, Optional

from llmfy.exception.llmfy_exception import LLMfyException

try:
    import faiss
except ImportError:
    faiss = None

try:
    import numpy as np
except ImportError:
    np = None


class FAISSIndex:
    def __init__(
        self,
        dim: int,
        index_type: Literal["flat", "hnsw", "lsh", "ivfflat", "ivfpq"] = "flat",
        nlist: int = 100,
        nprobe: int = 10,
        M: int = 32,
        ef_search: int = 50,
        nbits: int = 8,
        index: Optional[Any] = None,
    ):
        """
        Initialize FAISS Index.

        Args:
            dim (int): Dimension of vectors.

            index_type (str): Type of FAISS index (flat, ivfflat, ivfpq, hnsw, lsh).

            nlist (int): Number of clusters (for ivfflat and ivfpq).
                Higher nlist → finer partition, better recall, but more memory and slower training.
                Lower nlist → faster but worse recall.
                Depends on vectors? Yes → set relative to dataset size, depends on dataset size, larger datasets need higher nlist.
                Retraining? Yes. Changing nlist requires re-training and rebuilding index.
                Tunable at query time? ❌ No.
                Rule of thumb:
                    - nlist ≈ √N (where N is the number of vectors),
                    - Example: For 1M vectors → nlist ~ 1000.

            nprobe (int): Number of clusters to search (for ivfflat and ivfpq).
                Higher nprobe → better recall, slower search.
                Lower nprobe → faster, but may miss good neighbors.
                Depends on vectors? No → only affects search speed/recall.
                Retraining? No.
                Tunable at query time? ✅ Yes.
                Rule of thumb:
                    - Start with nprobe = 1–10% of nlist,
                    - Example: nlist = 1000 → nprobe = 10–50.

            M (int): Number of neighbors per node in the HNSW graph (for hnsw).
                Higher M → better recall, more memory, longer build time.
                Lower M → smaller memory, but lower recall.
                Depends on vectors? Not directly, but larger datasets usually benefit from higher M.
                Retraining? Yes → must rebuild graph if M changes.
                Tunable at query time? ❌ No.
                Rule of thumb:
                    - Typical range: 16–64,
                    - M = 32 is a common good default.

            ef_search (int): Size of candidate list during search (search depth) (for hnsw).
                Higher ef_search → better recall, slower query.
                Lower ef_search → faster, but may miss true neighbors.
                Depends on vectors? No.
                Retraining? No.
                Tunable at query time? ✅ Yes.
                Rule of thumb:
                    - Set ef_search around k * 2–10 (where k is top neighbors).
                    - Example: for k=10, ef_search=50 is common.

            nbits (int): Number of bits  per vector used for hashing (for lsh).
                Higher nbits → more fine-grained buckets, better recall, but more memory.
                Lower nbits → fewer buckets, faster, but worse recall.
                Depends on vectors? Yes → tied to dimension.
                Retraining? Yes → index must be rebuilt if nbits changes.
                Tunable at query time? ❌ No.
                Rule of thumb:
                    - Often nbits = 2 * dim or less.
                    - Example: for dim=128, try nbits=128 or nbits=256.


        """

        if faiss is None:
            raise LLMfyException(
                'faiss package is not installed. Install it using `pip install "llmfy[faiss-cpu]"`'
            )

        if np is None:
            raise LLMfyException(
                'numpy package is not installed. Install it using `pip install "llmfy[numpy]"`'
            )

        self.dim = dim
        self.index_type = index_type.lower()
        self.index = (
            self._create_index(dim, self.index_type, nlist, M, nbits)
            if not index
            else index
        )

        # IVF params
        self.nprobe = nprobe
        if "ivfflat" in self.index_type:
            self.index.nprobe = nprobe  # type: ignore

        # HNSW params
        self.ef_search = ef_search
        if "hnsw" in self.index_type:
            self.index.hnsw.efSearch = ef_search  # type: ignore

    def _create_index(
        self,
        dim: int,
        index_type: str,
        nlist: int,
        M: int,
        nbits: int,
    ):
        if faiss is None:
            raise LLMfyException(
                'faiss package is not installed. Install it using `pip install "llmfy[faiss-cpu]"`'
            )

        if np is None:
            raise LLMfyException(
                'numpy package is not installed. Install it using `pip install "llmfy[numpy]"`'
            )

        # Use IP (Inner Product / Cosine Similarity) best for semantic search, retrieval, RAG
        # with embeddings from models like Bedrock Titan, OpenAI, Cohere, etc.
        if index_type == "flat":
            return faiss.IndexFlatIP(dim)
        elif index_type == "hnsw":
            return faiss.IndexHNSWFlat(dim, M, faiss.METRIC_INNER_PRODUCT)
        elif index_type == "lsh":
            return faiss.IndexLSH(dim, nbits)
        elif index_type == "ivfflat":
            quantizer = faiss.IndexFlatIP(dim)
            return faiss.IndexIVFFlat(quantizer, dim, nlist, faiss.METRIC_INNER_PRODUCT)
        elif index_type == "ivfpq":
            quantizer = faiss.IndexFlatIP(dim)
            return faiss.IndexIVFPQ(
                quantizer, dim, nlist, M, nbits, faiss.METRIC_INNER_PRODUCT
            )
        else:
            raise ValueError(f"Unsupported index type: {index_type}")

    def train(self, vectors):
        """
        Train the index (required for IVF).

        Args:
            vectors (np.ndarray): _description_
        """
        self.index.train(vectors)  # type: ignore

    def add(self, vectors):
        """
        Add vectors to the index.

        Args:
            vectors (np.ndarray): _description_

        Raises:
            ValueError: _description_
        """
        if not self.index:
            raise ValueError("Index not initialized.")

        # Check train if is_trained is false
        if not self.index.is_trained:
            self.train(vectors)

        self.index.add(vectors)  # type: ignore

    def search(
        self,
        query,
        k: int = 5,
        nprobe: Optional[int] = None,
        ef_search: Optional[int] = None,
    ):
        """
        Search the index.

        Args:
            query (np.ndarray): _description_
            k (int, optional): _description_. Defaults to 5.
            nprobe (Optional[int], optional): _description_. Defaults to None.
            ef_search (Optional[int], optional): _description_. Defaults to None.

        Returns:
            distances (np.ndarray), indices (np.ndarray)
        """
        # Tuning HNSW params
        if "hnsw" in self.index_type:
            if ef_search:
                self.index.hnsw.efSearch = ef_search  # type: ignore
            else:
                self.index.hnsw.efSearch = self.ef_search  # type: ignore

        # Tuning IVF params
        if "ivfflat" in self.index_type:
            if nprobe:
                self.index.nprobe = nprobe  # type: ignore
            else:
                self.index.nprobe = self.nprobe  # type: ignore

        distances, indices = self.index.search(query, k)  # type: ignore
        return distances, indices

    def save(self, path: str):
        """
        Save the FAISS index to disk.
        """
        if faiss is None:
            raise LLMfyException(
                'faiss package is not installed. Install it using `pip install "llmfy[faiss-cpu]"`'
            )

        faiss.write_index(self.index, path)

    def load(self, path: str):
        """
        Load the FAISS index from disk.
        """
        if faiss is None:
            raise LLMfyException(
                'faiss package is not installed. Install it using `pip install "llmfy[faiss-cpu]"`'
            )

        self.index = faiss.read_index(path)
